@@ -22,6 +22,7 @@
 
 #include <pwd.h>
 #include <netdb.h>
+#include <assert.h>
 
 static const char toplevel_config_comment[] =
     " .notmuch-config - Configuration file for the notmuch mail system\n"
@@ -43,7 +44,13 @@ static const char new_config_comment[] =
     " The following options are supported here:\n"
     "\n"
     "\ttags	A list (separated by ';') of the tags that will be\n"
-    "\t	added to all messages incorporated by \"notmuch new\".\n";
+    "\t	added to all messages incorporated by \"notmuch new\".\n"
+    "\n"
+    "\tignore	A list (separated by ';') of file and directory names\n"
+    "\t	that will not be searched for messages by \"notmuch new\".\n"
+    "\n"
+    "\t	NOTE: *Every* file/directory that goes by one of those names will\n"
+    "\t	be ignored, independent of its depth/location in the mail store.\n";
 
 static const char user_config_comment[] =
     " User configuration\n"
@@ -83,6 +90,16 @@ static const char maildir_config_comment[] =
     "\tand update tags, while the \"notmuch tag\" and \"notmuch restore\"\n"
     "\tcommands will notice tag changes and update flags in filenames\n";
 
+static const char search_config_comment[] =
+    " Search configuration\n"
+    "\n"
+    " The following option is supported here:\n"
+    "\n"
+    "\texclude_tags\n"
+    "\t\tA ;-separated list of tags that will be excluded from\n"
+    "\t\tsearch results by default.  Using an excluded tag in a\n"
+    "\t\tquery will override that exclusion.\n";
+
 struct _notmuch_config {
     char *filename;
     GKeyFile *key_file;
@@ -94,7 +111,11 @@ struct _notmuch_config {
     size_t user_other_email_length;
     const char **new_tags;
     size_t new_tags_length;
+    const char **new_ignore;
+    size_t new_ignore_length;
     notmuch_bool_t maildir_synchronize_flags;
+    const char **search_exclude_tags;
+    size_t search_exclude_tags_length;
 };
 
 static int
@@ -220,6 +241,7 @@ notmuch_config_open (void *ctx,
     int file_had_new_group;
     int file_had_user_group;
     int file_had_maildir_group;
+    int file_had_search_group;
 
     if (is_new_ret)
 	*is_new_ret = 0;
@@ -250,7 +272,11 @@ notmuch_config_open (void *ctx,
     config->user_other_email_length = 0;
     config->new_tags = NULL;
     config->new_tags_length = 0;
+    config->new_ignore = NULL;
+    config->new_ignore_length = 0;
     config->maildir_synchronize_flags = TRUE;
+    config->search_exclude_tags = NULL;
+    config->search_exclude_tags_length = 0;
 
     if (! g_key_file_load_from_file (config->key_file,
 				     config->filename,
@@ -294,6 +320,7 @@ notmuch_config_open (void *ctx,
     file_had_new_group = g_key_file_has_group (config->key_file, "new");
     file_had_user_group = g_key_file_has_group (config->key_file, "user");
     file_had_maildir_group = g_key_file_has_group (config->key_file, "maildir");
+    file_had_search_group = g_key_file_has_group (config->key_file, "search");
 
 
     if (notmuch_config_get_database_path (config) == NULL) {
@@ -344,6 +371,19 @@ notmuch_config_open (void *ctx,
 	notmuch_config_set_new_tags (config, tags, 2);
     }
 
+    if (notmuch_config_get_new_ignore (config, &tmp) == NULL) {
+	notmuch_config_set_new_ignore (config, NULL, 0);
+    }
+
+    if (notmuch_config_get_search_exclude_tags (config, &tmp) == NULL) {
+	if (is_new) {
+	    const char *tags[] = { "deleted", "spam" };
+	    notmuch_config_set_search_exclude_tags (config, tags, 2);
+	} else {
+	    notmuch_config_set_search_exclude_tags (config, NULL, 0);
+	}
+    }
+
     error = NULL;
     config->maildir_synchronize_flags =
 	g_key_file_get_boolean (config->key_file,
@@ -386,6 +426,11 @@ notmuch_config_open (void *ctx,
 				maildir_config_comment, NULL);
     }
 
+    if (! file_had_search_group) {
+	g_key_file_set_comment (config->key_file, "search", NULL,
+				search_config_comment, NULL);
+    }
+
     if (is_new_ret)
 	*is_new_ret = is_new;
 
@@ -436,6 +481,48 @@ notmuch_config_save (notmuch_config_t *config)
     return 0;
 }
 
+static const char **
+_config_get_list (notmuch_config_t *config,
+		  const char *section, const char *key,
+		  const char ***outlist, size_t *list_length, size_t *ret_length)
+{
+    assert(outlist);
+
+    if (*outlist == NULL) {
+
+	char **inlist = g_key_file_get_string_list (config->key_file,
+					     section, key, list_length, NULL);
+	if (inlist) {
+	    unsigned int i;
+
+	    *outlist = talloc_size (config, sizeof (char *) * (*list_length + 1));
+
+	    for (i = 0; i < *list_length; i++)
+		(*outlist)[i] = talloc_strdup (*outlist, inlist[i]);
+
+	    (*outlist)[i] = NULL;
+
+	    g_strfreev (inlist);
+	}
+    }
+
+    if (ret_length)
+	*ret_length = *list_length;
+
+    return *outlist;
+}
+
+static void
+_config_set_list (notmuch_config_t *config,
+		  const char *group, const char *name,
+		  const char *list[],
+		  size_t length, const char ***config_var )
+{
+    g_key_file_set_string_list (config->key_file, group, name, list, length);
+    talloc_free (*config_var);
+    *config_var = NULL;
+}
+
 const char *
 notmuch_config_get_database_path (notmuch_config_t *config)
 {
@@ -521,91 +608,71 @@ notmuch_config_set_user_primary_email (notmuch_config_t *config,
 }
 
 const char **
-notmuch_config_get_user_other_email (notmuch_config_t *config,
-				     size_t *length)
+notmuch_config_get_user_other_email (notmuch_config_t *config,   size_t *length)
 {
-    char **emails;
-    size_t emails_length;
-    unsigned int i;
+    return _config_get_list (config, "user", "other_email",
+			     &(config->user_other_email),
+			     &(config->user_other_email_length), length);
+}
 
-    if (config->user_other_email == NULL) {
-	emails = g_key_file_get_string_list (config->key_file,
-					     "user", "other_email",
-					     &emails_length, NULL);
-	if (emails) {
-	    config->user_other_email = talloc_size (config,
-						    sizeof (char *) *
-						    (emails_length + 1));
-	    for (i = 0; i < emails_length; i++)
-		config->user_other_email[i] = talloc_strdup (config->user_other_email,
-							     emails[i]);
-	    config->user_other_email[i] = NULL;
+const char **
+notmuch_config_get_new_tags (notmuch_config_t *config,   size_t *length)
+{
+    return _config_get_list (config, "new", "tags",
+			     &(config->new_tags),
+			     &(config->new_tags_length), length);
+}
 
-	    g_strfreev (emails);
-
-	    config->user_other_email_length = emails_length;
-	}
-    }
-
-    *length = config->user_other_email_length;
-    return config->user_other_email;
+const char **
+notmuch_config_get_new_ignore (notmuch_config_t *config, size_t *length)
+{
+    return _config_get_list (config, "new", "ignore",
+			     &(config->new_ignore),
+			     &(config->new_ignore_length), length);
 }
 
 void
 notmuch_config_set_user_other_email (notmuch_config_t *config,
-				     const char *other_email[],
+				     const char *list[],
 				     size_t length)
 {
-    g_key_file_set_string_list (config->key_file,
-				"user", "other_email",
-				other_email, length);
-
-    talloc_free (config->user_other_email);
-    config->user_other_email = NULL;
-}
-
-const char **
-notmuch_config_get_new_tags (notmuch_config_t *config,
-			     size_t *length)
-{
-    char **tags;
-    size_t tags_length;
-    unsigned int i;
-
-    if (config->new_tags == NULL) {
-	tags = g_key_file_get_string_list (config->key_file,
-					   "new", "tags",
-					   &tags_length, NULL);
-	if (tags) {
-	    config->new_tags = talloc_size (config,
-					    sizeof (char *) *
-					    (tags_length + 1));
-	    for (i = 0; i < tags_length; i++)
-		config->new_tags[i] = talloc_strdup (config->new_tags,
-						     tags[i]);
-	    config->new_tags[i] = NULL;
-
-	    g_strfreev (tags);
-
-	    config->new_tags_length = tags_length;
-	}
-    }
-
-    *length = config->new_tags_length;
-    return config->new_tags;
+    _config_set_list (config, "user", "other_email", list, length,
+		     &(config->user_other_email));
 }
 
 void
 notmuch_config_set_new_tags (notmuch_config_t *config,
-			     const char *new_tags[],
-			     size_t length)
+				     const char *list[],
+				     size_t length)
 {
-    g_key_file_set_string_list (config->key_file,
-				"new", "tags",
-				new_tags, length);
+    _config_set_list (config, "new", "tags", list, length,
+		     &(config->new_tags));
+}
 
-    talloc_free (config->new_tags);
-    config->new_tags = NULL;
+void
+notmuch_config_set_new_ignore (notmuch_config_t *config,
+			       const char *list[],
+			       size_t length)
+{
+    _config_set_list (config, "new", "ignore", list, length,
+		     &(config->new_ignore));
+}
+
+const char **
+notmuch_config_get_search_exclude_tags (notmuch_config_t *config, size_t *length)
+{
+    return _config_get_list (config, "search", "exclude_tags",
+			     &(config->search_exclude_tags),
+			     &(config->search_exclude_tags_length), length);
+}
+
+void
+notmuch_config_set_search_exclude_tags (notmuch_config_t *config,
+				      const char *list[],
+				      size_t length)
+{
+    _config_set_list (config, "search", "exclude_tags", list, length,
+		      &(config->search_exclude_tags));
 }
 
 /* Given a configuration item of the form <group>.<key> return the
@@ -684,7 +751,7 @@ notmuch_config_command_get (void *ctx, char *item)
 	for (i = 0; i < length; i++)
 	    printf ("%s\n", value[i]);
 
-	free (value);
+	g_strfreev (value);
     }
 
     notmuch_config_close (config);
@@ -732,20 +799,78 @@ notmuch_config_command_set (void *ctx, char *item, int argc, char *argv[])
     return ret;
 }
 
+static int
+notmuch_config_command_list (void *ctx)
+{
+    notmuch_config_t *config;
+    char **groups;
+    size_t g, groups_length;
+
+    config = notmuch_config_open (ctx, NULL, NULL);
+    if (config == NULL)
+	return 1;
+
+    groups = g_key_file_get_groups (config->key_file, &groups_length);
+    if (groups == NULL)
+	return 1;
+
+    for (g = 0; g < groups_length; g++) {
+	char **keys;
+	size_t k, keys_length;
+
+	keys = g_key_file_get_keys (config->key_file,
+				    groups[g], &keys_length, NULL);
+	if (keys == NULL)
+	    continue;
+
+	for (k = 0; k < keys_length; k++) {
+	    char *value;
+
+	    value = g_key_file_get_string (config->key_file,
+					   groups[g], keys[k], NULL);
+	    if (value != NULL) {
+		printf ("%s.%s=%s\n", groups[g], keys[k], value);
+		free (value);
+	    }
+	}
+
+	g_strfreev (keys);
+    }
+
+    g_strfreev (groups);
+
+    notmuch_config_close (config);
+
+    return 0;
+}
+
 int
 notmuch_config_command (void *ctx, int argc, char *argv[])
 {
     argc--; argv++; /* skip subcommand argument */
 
-    if (argc < 2) {
-	fprintf (stderr, "Error: notmuch config requires at least two arguments.\n");
+    if (argc < 1) {
+	fprintf (stderr, "Error: notmuch config requires at least one argument.\n");
 	return 1;
     }
 
-    if (strcmp (argv[0], "get") == 0)
+    if (strcmp (argv[0], "get") == 0) {
+	if (argc != 2) {
+	    fprintf (stderr, "Error: notmuch config get requires exactly "
+		     "one argument.\n");
+	    return 1;
+	}
 	return notmuch_config_command_get (ctx, argv[1]);
-    else if (strcmp (argv[0], "set") == 0)
+    } else if (strcmp (argv[0], "set") == 0) {
+	if (argc < 2) {
+	    fprintf (stderr, "Error: notmuch config set requires at least "
+		     "one argument.\n");
+	    return 1;
+	}
 	return notmuch_config_command_set (ctx, argv[1], argc - 2, argv + 2);
+    } else if (strcmp (argv[0], "list") == 0) {
+	return notmuch_config_command_list (ctx);
+    }
 
     fprintf (stderr, "Unrecognized argument for notmuch config: %s\n",
 	     argv[0]);
